"""
Implementation of the matching_anchors_filter script
"""
from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Optional, Sequence, TypedDict, Union, cast

import configpile as cp
import numpy as np
import tybles as tb
from filelock import FileLock
from numpy.typing import NDArray
from typing_extensions import Annotated

from ..lib.data import LoggingLevel, NameRow, PathPattern, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.math import local_max, make_continuum
from ..lib.util import log_task_name_and_time
from ..rassine.types import RassineBasicOutput, RassineParameters, RassinePickle


@dataclass(frozen=True)
class MatchingAnchorsRow:
    """
    Describes the effect of the matching_anchors step
    """

    name: str
    is_master: bool
    n_anchors_before: int
    n_anchors_ooc: int
    n_anchors_twin: int
    n_anchors_adding: int

    @staticmethod
    def schema() -> tb.Schema[MatchingAnchorsRow]:
        return tb.schema(
            MatchingAnchorsRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


class MasterToolPickle(TypedDict):
    """Format of the master tool created by the matching_anchors_scan step"""

    curve: NDArray[np.bool_]
    border: NDArray[np.float64]
    wave: NDArray[np.float64]
    threshold: float
    tolerance: float
    fraction: float
    nb_copies_master: int

    #: Warning: only the final path component of the master file is included
    master_filename: Optional[str]


class AnchorParameters(TypedDict):
    #: Master tool file used to compute the matching anchors
    master_tool: str
    master_filename: Optional[str]
    threshold: float
    tolerance: float
    fraction: float
    nb_copies_master: int
    sub_dico_used: str


class AnchorOutput(TypedDict):
    parameters: AnchorParameters
    continuum_linear: NDArray[np.float64]
    anchor_wave: NDArray[np.float64]
    anchor_flux: NDArray[np.float64]
    anchor_index: NDArray[np.int64]


class AnchorPickle(TypedDict):
    #: Wavelength
    wave: NDArray[np.float64]
    #: Initial flux
    flux: NDArray[np.float64]
    #: Initial flux error, passed through by RASSINE
    flux_err: Optional[NDArray[np.float64]]
    #: Smoothed flux
    flux_used: NDArray[np.float64]
    #: Rassine output continuum
    output: RassineBasicOutput
    #: Rassine derived parameters
    parameters: RassineParameters

    matching_anchors: AnchorOutput


class MatchingDiffParameters(TypedDict):
    reference_continuum: str
    savgol_window: int
    recenter: bool
    sub_dico_used: str  # always matching_nachors


class MatchingDiffOutput(TypedDict):
    parameters: MatchingDiffParameters
    continuum_linear: NDArray[np.float64]


class MatchingPickle(TypedDict):
    #: Wavelength
    wave: NDArray[np.float64]
    #: Initial flux
    flux: NDArray[np.float64]
    #: Initial flux error, passed through by RASSINE
    flux_err: Optional[NDArray[np.float64]]
    #: Smoothed flux
    flux_used: NDArray[np.float64]
    #: Rassine output continuum
    output: RassineBasicOutput
    #: Rassine derived parameters
    parameters: RassineParameters

    matching_anchors: AnchorOutput
    matching_diff: MatchingDiffOutput



"""Implements the matching_diff script"""
import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Sequence, TypedDict

import configpile as cp
import numpy as np
from numpy.typing import NDArray
from typing_extensions import Annotated

from ..lib.analysis import rolling_iq
from ..lib.data import LoggingLevel, NameRow, PathPattern, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.math import smooth
from ..lib.util import log_task_name_and_time
from ..rassine.types import RassineBasicOutput, RassineParameters
from .types import (
    AnchorOutput,
    AnchorPickle,
    MatchingDiffOutput,
    MatchingDiffParameters,
    MatchingPickle,
)


def matching_diff_continuum(
    input_data: AnchorPickle,
    master_name: str,
    master_data: AnchorPickle,
    savgol_window: int = 200,
    zero_point: bool = False,
) -> MatchingPickle:
    """
    Match the continuum of individual spectra to a reference spectrum with a savgol filtering on the spectra difference.

    Creates the matching_diff dictionary

    Args:
        path: Path of the pickle file to process
        master_: Name of the RASSINE master spectrum file.
        savgol window: Length of the window for the savgol filtering.
        zero_point: No more used ?
    """
    sub_dico = "matching_anchors"
    dx = np.diff(master_data["wave"])[0]
    length_clip = int(100 / dx)  # smoothing on 100 \ang for the tellurics clean

    spectre = input_data["flux_used"]

    par: MatchingDiffParameters = {
        "reference_continuum": master_name,
        "savgol_window": savgol_window,
        "recenter": zero_point,
        "sub_dico_used": sub_dico,
    }

    cont = input_data[sub_dico]["continuum_linear"]
    cont2 = master_data["flux_used"] / master_data[sub_dico]["continuum_linear"]

    cont1 = spectre / cont
    diff = cont1 - cont2
    med_value = np.nanmedian(diff)
    for _ in range(3):  # needed to avoid artefact induced by tellurics
        q1, q3, iq = rolling_iq(diff, window=length_clip)
        diff[(diff > q3 + 1.5 * iq) | (diff < q1 - 1.5 * iq)] = med_value
        diff[diff < q1 - 1.5 * iq] = med_value

    correction = smooth(diff, savgol_window, shape="savgol")
    correction = smooth(correction, savgol_window, shape="savgol")
    if zero_point:
        correction = correction - np.nanmedian(correction)
    cont_corr: NDArray[np.float64] = cont.copy() / (1 - correction.copy())
    this_output: MatchingDiffOutput = {"parameters": par, "continuum_linear": cont_corr}
    return {
        "wave": input_data["wave"],
        "flux": input_data["flux"],
        "flux_err": input_data["flux_err"],
        "flux_used": input_data["flux_used"],
        "output": input_data["output"],
        "parameters": input_data["parameters"],
        "matching_anchors": input_data["matching_anchors"],
        "matching_diff": this_output,
    }


@dataclass(frozen=True)
class Task(cp.Config):
    """
    SAVGOL step

    Match the continuum of individual spectra to a reference spectrum
    with a savgol filtering on the spectra difference.

    Creates the matching_diff dictionary
    """

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #: Table of stacked spectra from which to read the individual spectrum names
    #:
    #: The parameters process_table, process_pattern and process_indices must be all provided together
    process_table: Annotated[
        Path,
        cp.Param.store(
            cp.parsers.path_parser.empty_means_none(), short_flag_name="-I", default_value=""
        ),
    ]

    #: Relative path (pattern) for the files to process
    #:
    #: The parameters process_table, process_pattern and process_indices must be all provided together
    process_pattern: Annotated[
        PathPattern,
        cp.Param.store(
            PathPattern.parser().empty_means_none(), short_flag_name="-i", default_value=""
        ),
    ]

    #: Indices of spectrum to process
    #:
    #: The parameters process_table, process_pattern and process_indices must be all provided together
    process_indices: Annotated[
        Sequence[int],
        cp.Param.append1(
            cp.parsers.int_parser,
            positional=cp.Positional.ONE_OR_MORE,
            long_flag_name=None,
            short_flag_name=None,
        ),
    ]

    #: Length of the window for SAVGOL filtering
    savgol_window: Annotated[int, cp.Param.store(cp.parsers.int_parser, default_value="200")]

    #: Processed master spectrum / Anchor file
    anchor_file: Annotated[Path, cp.Param.store(cp.parsers.path_parser)]

    #: TODO: document
    zero_point: Annotated[bool, cp.Param.store(cp.parsers.bool_parser, default_value="false")]


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    t.pickle_protocol.set()

    master_path = t.root / t.anchor_file
    master_name = master_path.name
    master_data = open_pickle(master_path, AnchorPickle)

    logging.info("Processing %d stacked spectra provided", len(t.process_indices))
    tyble = NameRow.schema().read_csv(t.root / t.process_table, return_type="Tyble")
    for i in t.process_indices:
        file = t.process_pattern.to_path(t.root, tyble[i].name)
        input_data = open_pickle(file, AnchorPickle)
        output_data = matching_diff_continuum(
            input_data=input_data,
            master_name=master_name,
            master_data=master_data,
            savgol_window=t.savgol_window,
            zero_point=t.zero_point,
        )
        save_pickle(file, output_data)


def cli():
    run(Task.from_command_line_())



import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, List, Optional, Sequence

import configpile as cp
import numpy as np
from numpy.typing import NDArray
from typing_extensions import Annotated

from ..lib.analysis import grouping
from ..lib.data import LoggingLevel, PathPattern, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.util import log_task_name_and_time
from ..rassine.types import RassinePickle
from ..stacking.types import StackedBasicRow
from .types import MasterToolPickle


def intersect_all_continuum_sphinx(
    names: Sequence[Path],
    output_file: Path,
    master_spectrum: Optional[Path] = None,
    copies_master: int = 0,
    fraction: float = 0.2,
    threshold: float = 0.66,
    tolerance: float = 0.5,
):
    """
    Search for the intersection of all the anchors points in a list of filename and update the
    selection of anchor points in the same files.

    For each anchor point the fraction of the closest distance to a neighborhood is used.
    Anchor point under the threshold are removed (1 = most deleted, 0 = most kept).
    Possible to use multiprocessing with nthreads cpu.
    If you want to fix the anchors points, enter a master spectrum path and the number of copies
    you want of it.

    Args:
        names: List of RASSINE files.
        output_file: Master tool file to write
        master_spectrum: Name of the RASSINE master spectrum file.
        copies_master: Number of copies of the master.

                       If 0 value is specified, copies_master is set to 2*N
                       with N the number of RASSINE files.
        fraction: Parameter of the model between 0 and 1
        treshold: Parameter of the model between 0 and 1
        tolerance: Parameter of the model between 0 and 1
    """

    # TODO: master_spectrum and copies_master is the addition of the master to
    # add some importance for the master data to the process

    logging.info("Loading of the files, wait ...")

    # removal of other Master_tool in Bash script

    # TOCHECK: removed
    # previous_files = glob.glob(directory + "Master_tool*.p")
    # for file_to_delete in previous_files:
    #     os.system("rm " + file_to_delete)

    wave = open_pickle(names[0], RassinePickle)["wave"]
    save: List[NDArray[np.int64]] = [
        open_pickle(filename, RassinePickle)["output"]["anchor_index"] for filename in names
    ]

    if master_spectrum is not None:
        if copies_master == 0:
            logging.warning(
                "You have to specify the number of copy of the master file you want as argument."
            )
            copies_master = 2 * len(names)

            logging.warning("Default value of master copies fixed at %d", copies_master)
        file: RassinePickle = open_pickle(master_spectrum, RassinePickle)
        for j in range(copies_master):
            save.append(file["output"]["anchor_index"])
    logging.info("Got %d files", len(save))
    sum_mask = []
    all_idx = np.hstack(save)

    logging.info("Computation of the intersection of the anchors points, wait ...")
    for j in range(len(names)):
        # plt.scatter(save[j],j*np.ones(len(save[j])))
        diff = np.min([np.diff(save[j][1:]), np.diff(save[j][0:-1])], axis=0)
        diff = np.array([diff[0]] + list(diff) + [diff[-1]])
        diff = diff * fraction
        diff = diff.astype("int")
        mask = np.zeros(len(wave))
        new = []
        for k in range(len(save[j])):
            new.append(save[j][k] + np.arange(-diff[k], diff[k]))
        new = np.unique(np.hstack(new))
        new = new[(new > 0) & (new < len(mask))]
        mask[new.astype("int")] = 1
        sum_mask.append(mask)
    sum_mask = np.array(sum_mask)
    sum_mask_vert = np.sum(sum_mask, axis=0)

    strat = np.linspace(int(sum_mask_vert.min()), int(sum_mask_vert.max()), 10).astype("int")
    strat = strat[strat != strat[1]]  # suppress the first level

    for j in range(len(strat) - 1)[::-1]:
        sum_mask_vert[(sum_mask_vert >= strat[j]) & (sum_mask_vert < strat[j + 1])] = strat[j]
    for j in range(len(strat[0:-1]))[::-1]:
        sum_mask_vert[sum_mask_vert == strat[0:-1][j]] = strat[j + 1]

    # sum_mask_vert -= np.diff(strat)[0]
    sum_mask_vert[sum_mask_vert == np.unique(sum_mask_vert)[0]] = 0

    mask_vert2 = sum_mask_vert.copy()
    mask_vert2[0] = 0
    mask_vert2[-1] = 0

    for j in range(len(mask_vert2) - 2):
        j += 1
        if (mask_vert2[j] > mask_vert2[j - 1]) | (mask_vert2[j] > mask_vert2[j + 1]):
            # supression of delta peak (not possible because all the peaks cannot be situated
            # #at the exact same wavelength), allow the use of grouping function here after
            mask_vert2[j] = np.max([mask_vert2[j - 1], mask_vert2[j + 1]])

    val, border = grouping(mask_vert2, 1, 1)
    border = np.hstack([border, np.array([i[0] for i in val])[:, np.newaxis]])

    null = np.where(border[:, -1] == 0)[0]

    area = []
    small_length = []
    small_center = []
    for j in range(len(null) - 1):
        area.append(
            np.sum(border[null[j] + 1 : null[j + 1], 2] * border[null[j] + 1 : null[j + 1], 3])
        )
        peak = np.where(
            border[null[j] + 1 : null[j + 1], 3] == (border[null[j] + 1 : null[j + 1], 3].max())
        )[0]
        peak = peak[border[null[j] + 1 : null[j + 1], 2][peak].argmax()]
        small_length.append(border[null[j] + 1 : null[j + 1], 2][peak])
        small_center.append(border[null[j] + 1 : null[j + 1], 0][peak] + small_length[-1] / 2)

    left = border[null, 1][0:-1]
    right = border[null, 0][1:]

    center2_: List[int] = []
    for i, j in zip(left, right):
        slc = all_idx[(all_idx >= i) & (all_idx <= j)]
        if len(slc):
            center2_.append(int(np.median(slc)))
        else:
            center2_.append(0)
    center2 = np.array(center2_, dtype="int64")  # kept
    center2[center2 < 0] = 0

    large_length = right - left
    large_center = left + large_length / 2

    center = np.mean(np.array([[small_center], [large_center]]), axis=0)[0]
    windows = np.mean(np.array([[small_length], [large_length]]), axis=0)[0]
    height = area / windows

    center = np.round(center, 0).astype("int")
    windows = np.round(windows / 2, 0).astype("int")
    height = np.round(height, 0).astype("int")

    center = center2

    # take any scan
    gri = wave

    stack_vert3 = np.zeros(len(sum_mask_vert))

    stack_vert = np.zeros(len(sum_mask_vert))
    for j in range(len(save)):  # pylint: disable=C0200
        stack_vert[save[j]] += 1

    for j in range(len(center)):  # pylint: disable=C0200
        if center[j] != np.nan:
            stack_vert3[center[j] - windows[j] : center[j] + windows[j] + 1] = height[j]

    stack_vert3[stack_vert3 < (len(names) * threshold)] = 0
    stack_vert3[stack_vert3 > (len(names))] = len(names)
    val, border = grouping(stack_vert3, 1, 1)
    border = np.hstack([border, np.array([i[0] for i in val])[:, np.newaxis]])

    for j in range(len(border)):
        if np.sum(stack_vert[int(border[j, 0]) : int(border[j, 1]) + 1]) < (
            border[j, 3] * tolerance
        ):
            stack_vert3[int(border[j, 0]) : int(border[j, 1]) + 1] = 0

    liste_ = []
    for j in range(len(stack_vert3) - 2):
        j += 1
        if (stack_vert3[j] != stack_vert3[j - 1]) | (stack_vert3[j] != stack_vert3[j + 1]):
            liste_.append(j + 1)
    liste = np.array(liste_)
    if len(liste) != 0:
        stack_vert3[liste] = 0

    val, border = grouping(stack_vert3, 1, 1)
    border = np.hstack([border, np.array([i[0] for i in val])[:, np.newaxis]])

    val2, border2 = grouping(stack_vert3, 1, 1)
    border2 = np.hstack([border2, np.array([i[0] for i in val2])[:, np.newaxis]])
    border2 = border2[border2[:, -1] > 0]
    curve = stack_vert3 > 0

    output_cluster: MasterToolPickle = {
        "curve": curve,
        "border": border2,
        "wave": gri,
        "threshold": threshold,
        "tolerance": tolerance,
        "fraction": fraction,
        "nb_copies_master": copies_master,
        "master_filename": None if master_spectrum is None else master_spectrum.name,
    }

    save_pickle(output_file, output_cluster)


@dataclass(frozen=True)
class Task(cp.Config):
    """
    Intersection tool

    Searches for the intersection of all the anchors points in a list of filenames and updates
    the selection of anchor points in the same files.

    For each anchor point the fraction of the closest distance to a neighbourhood is used.
    Anchor point under the threshold are removed (1 = most deleted, 0 = most kept).
    Possible to use multiprocessing with nthreads cpu.
    If you want to fix the anchors points, enter a master spectrum path and the number of
    copies you want of it.

    It reads the files matching RASSINE*.p in the given input directory.
    """

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #: Table of spectra from which to read the individual file names (produced by stacking step)
    input_table: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, short_flag_name="-I"),
    ]

    #: Path pattern for input files
    input_pattern: Annotated[
        PathPattern,
        cp.Param.store(PathPattern.parser(), short_flag_name="-i", default_value="RASSINE_{}.p"),
    ]

    #: Master tool output file
    output_file: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-o")]

    #: Name of the RASSINE master spectrum file
    master_spectrum: Annotated[
        Optional[Path], cp.Param.store(cp.parsers.path_parser.empty_means_none(), default_value="")
    ]

    #: Do not use a RASSINE master spectrum file
    no_master_spectrum: ClassVar[cp.Expander] = cp.Expander.make(
        new_flag="--master-spectrum", new_value=""
    )

    #: Number of copy of the master.
    #:
    #: If value 0 is specified, copies_master is set to 2*N with N the number of RASSINE files.
    copies_master: Annotated[int, cp.Param.store(cp.parsers.int_parser, default_value="0")]

    #: Unused: number of threads for multiprocessing
    nthreads: Annotated[int, cp.Param.store(cp.parsers.int_parser, default_value="1")]

    #: Parameter of the model between 0 and 1
    fraction: Annotated[float, cp.Param.store(cp.parsers.float_parser, default_value="0.2")]

    #: Parameter of the model between 0 and 1
    threshold: Annotated[float, cp.Param.store(cp.parsers.float_parser, default_value="0.65")]

    #: Parameter of the model between 0 and 1
    tolerance: Annotated[float, cp.Param.store(cp.parsers.float_parser, default_value="0.5")]


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    t.pickle_protocol.set()

    input_tyble = StackedBasicRow.schema().read_csv(t.root / t.input_table, return_type="Tyble")
    files = [t.input_pattern.to_path(t.root, r.name) for r in input_tyble]

    intersect_all_continuum_sphinx(
        files,
        master_spectrum=(None if t.master_spectrum is None else t.root / t.master_spectrum),
        copies_master=t.copies_master,
        fraction=t.fraction,
        threshold=t.threshold,
        tolerance=t.tolerance,
        output_file=t.root / t.output_file,
    )


def cli():
    run(Task.from_command_line_())



from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Sequence, TypedDict

import configpile as cp
import numpy as np
import numpy.typing as npt
import tybles as tb
from filelock import FileLock
from numpy.typing import NDArray
from scipy.interpolate import interp1d
from typing_extensions import Annotated

from ..lib.analysis import find_nearest1
from ..lib.data import LoggingLevel, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.math import absurd_minus_99_9, create_grid, doppler_r
from ..lib.util import log_task_name_and_time
from .types import (
    IndividualImportedRow,
    IndividualReinterpolatedRow,
    PickledIndividualSpectrum,
    ReinterpolatedSpectrumPickle,
)


@dataclass(frozen=True)
class Task(cp.Config):
    """Reinterpolates the spectra to match the stellar frame"""

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #

    prog_ = Path(__file__).stem

    ini_strict_sections_ = [Path(__file__).stem.split("_")[0]]

    #: Input spectrum table
    input_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-I")]

    #: Output spectrum table
    output_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-O")]

    #: Relative path to the folder containing the spectra
    input_folder: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-i")]

    #: Name of the output directory. If None, the output directory is created at the same location than the spectra.
    output_folder: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-o")]

    #: Indices of spectrum to process
    #:
    #: If not provided, all spectra are processed
    inputs: Annotated[
        Sequence[int],
        cp.Param.append1(
            cp.parsers.int_parser,
            positional=cp.Positional.ZERO_OR_MORE,
            long_flag_name=None,
            short_flag_name=None,
        ),
    ]

    #: Wavelength step in angstrom used to produce the equidistant wavelength vector
    dlambda: Annotated[float, cp.Param.store(cp.parsers.float_parser)]

    def validate_output_folder(self) -> Optional[cp.Err]:
        return cp.Err.check(
            (self.root / self.output_folder).is_dir(), "The output directory needs to exist"
        )


@dataclass(frozen=True)
class ReinterpolationSettings:
    wave_min_k: np.float64
    wave_max_k: np.float64
    hole_left_k: np.float64
    hole_right_k: np.float64
    dlambda: np.float64
    nb_bins: int
    static_grid: NDArray[np.float64]
    wave_ref: int

    @staticmethod
    def make(
        wave_min_k: np.float64,
        wave_max_k: np.float64,
        hole_left_k: np.float64,
        hole_right_k: np.float64,
        dlambda: np.float64,
        nb_bins: int,
    ) -> ReinterpolationSettings:
        static_grid = create_grid(wave_min_k, dlambda, nb_bins)
        wave_ref = find_nearest1(static_grid, 5500)[0]
        return ReinterpolationSettings(
            wave_min_k=wave_min_k,
            wave_max_k=wave_max_k,
            hole_left_k=hole_left_k,
            hole_right_k=hole_right_k,
            dlambda=dlambda,
            nb_bins=nb_bins,
            static_grid=static_grid,
            wave_ref=wave_ref,
        )

    @staticmethod
    def from_spectrum_data(
        wave_min: NDArray[np.float64],
        wave_max: NDArray[np.float64],
        hole_left: NDArray[np.float64],
        hole_right: NDArray[np.float64],
        dlambda: np.float64,
    ) -> ReinterpolationSettings:
        hole_left = hole_left[hole_left != absurd_minus_99_9]
        hole_right = hole_right[hole_right != absurd_minus_99_9]

        if len(hole_left) != 0 and len(hole_right) != 0:
            hole_left_k = np.min(hole_left) - 0.5  # increase of 0.5 the gap limit by security
            hole_right_k = np.max(hole_right) + 0.5  # increase of 0.5 the gap limit by security
        else:
            hole_left_k = absurd_minus_99_9
            hole_right_k = absurd_minus_99_9
        wave_min_k = wave_min.max()
        wave_max_k = wave_max.min()
        nb_bins: int = int(np.ceil((wave_max_k - wave_min_k) / dlambda))
        return ReinterpolationSettings.make(
            wave_min_k=wave_min_k,
            wave_max_k=wave_max_k,
            hole_left_k=np.float64(hole_left_k),
            hole_right_k=np.float64(hole_right_k),
            dlambda=dlambda,
            nb_bins=nb_bins,
        )


def reinterpolate(
    t: Task, row: IndividualImportedRow, s: ReinterpolationSettings
) -> IndividualReinterpolatedRow:
    input_file = t.root / t.input_folder / (row.name + ".p")
    output_file = t.root / t.output_folder / (row.name + ".p")
    spectrum = open_pickle(input_file, PickledIndividualSpectrum)

    # raw spectrum
    wave = spectrum["wave"]
    flux = spectrum["flux"]
    flux_err = spectrum["flux_err"]

    # reinterpolated using static_grid
    static_grid = s.static_grid
    new_flux = interp1d(
        doppler_r(wave, row.rv_shift)[1],
        flux,
        kind="cubic",
        bounds_error=False,
        fill_value="extrapolate",
    )(static_grid)
    new_flux_err = interp1d(
        doppler_r(wave, row.rv_shift)[1],
        flux_err,
        kind="linear",
        bounds_error=False,
        fill_value="extrapolate",
    )(static_grid)

    mask2 = (static_grid >= (s.hole_left_k - s.dlambda / 2.0)) & (
        static_grid <= (s.hole_right_k + s.dlambda / 2.0)
    )
    new_flux[mask2] = 0
    new_flux_err[mask2] = 1

    continuum_5500 = np.nanpercentile(new_flux[s.wave_ref - 50 : s.wave_ref + 50], 95)
    SNR = np.sqrt(continuum_5500)

    out: ReinterpolatedSpectrumPickle = {
        "flux": new_flux,
        "flux_err": new_flux_err,
        "RV_sys": row.rv_mean,
        "RV_shift": row.rv_shift,
        "SNR_5500": SNR,
        "berv": row.berv,
        "lamp_offset": row.lamp_offset,
        "plx_mas": row.plx_mas,
        "acc_sec": row.acc_sec,
        "instrument": row.instrument,
        "mjd": row.mjd,
        "jdb": row.jdb,
        "hole_left": s.hole_left_k,
        "hole_right": s.hole_right_k,
        "wave_min": s.wave_min_k,
        "wave_max": s.wave_max_k,
        "dwave": s.dlambda,
    }
    save_pickle(output_file, out)
    return IndividualReinterpolatedRow(
        name=row.name,
        instrument=row.instrument,
        mjd=row.mjd,
        jdb=row.jdb,
        model=row.model,
        rv_mean=row.rv_mean,
        rv_shift=row.rv_shift,
        SNR_5500=SNR,
        berv=row.berv,
        lamp_offset=row.lamp_offset,
        plx_mas=row.plx_mas,
        acc_sec=row.acc_sec,
        wave_min=s.wave_min_k,
        wave_max=s.wave_max_k,
        dwave=s.dlambda,
        nb_bins=np.int64(s.nb_bins),
        hole_left=s.hole_left_k,
        hole_right=s.hole_right_k,
        vrad=row.vrad,
        svrad=row.svrad,
        drift=row.drift,
    )


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    t.pickle_protocol.set()
    (t.root / t.output_folder).mkdir(parents=True, exist_ok=True)
    (t.root / t.output_table).parent.mkdir(parents=True, exist_ok=True)

    logging.debug(f"Reading {t.root/t.input_table}")
    tyble = IndividualImportedRow.schema().read_csv(t.root / t.input_table, return_type="Tyble")
    df = tyble.data_frame

    # compute hole boundaries
    settings = ReinterpolationSettings.from_spectrum_data(
        wave_min=np.round(df["wave_min"].to_numpy(), 8),  # to take into account float32
        wave_max=np.round(df["wave_max"].to_numpy(), 8),  # to take into account float32
        hole_left=df["hole_left"].to_numpy(),
        hole_right=df["hole_right"].to_numpy(),
        dlambda=np.float64(t.dlambda),
    )

    inputs: Sequence[int] = t.inputs
    if not inputs:
        inputs = list(range(len(df)))

    output_rows: List[IndividualReinterpolatedRow] = []
    for i in inputs:
        r = tyble[i]
        output_rows.append(reinterpolate(t, r, settings))

    output_table = t.root / t.output_table
    output_table_lockfile = output_table.with_suffix(output_table.suffix + ".lock")

    logging.debug(f"Appending to output table {output_table}")
    with FileLock(output_table_lockfile):
        df = IndividualReinterpolatedRow.schema().from_rows(output_rows, return_type="DataFrame")
        df.to_csv(output_table, header=not output_table.exists(), mode="a", index=False)


def cli() -> None:
    run(Task.from_command_line_())



from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

import numpy as np
import tybles as tb
from numpy.typing import NDArray


@dataclass(frozen=True)
class DACE:
    """Columns of the DACE table we import"""

    #: Full filename including folders, folders may be outdated
    fileroot: str

    #: Observation date/time in MJD
    mjd: np.float64

    #: Optional RV shift correction km/s
    model: np.float64

    @staticmethod
    def schema() -> tb.Schema[DACE]:
        return tb.schema(DACE, order_columns=True, missing_columns="error", extra_columns="drop")

    #: Radial velocity
    vrad: np.float64

    #: Radial velocity uncertainty
    svrad: np.float64

    #: Instrumental drift
    drift_used: np.float64


@dataclass(frozen=True)
class IndividualBasicRow:
    """Columns of the DACE data we extract for the rest of the processing"""

    #: Spectrum name without path and extension
    name: str

    #: Raw filename
    raw_filename: str

    #: Observation date/time in MJD
    mjd: np.float64

    #: Optional RV shift correction in km/s
    model: np.float64

    #: Median value of model (same for all spectra) in km/s
    rv_mean: np.float64

    #: Difference model - rv_mean in km/s
    rv_shift: np.float64

    #: Radial velocity
    vrad: np.float64

    #: Radial velocity uncertainty
    svrad: np.float64

    #: Instrumental drift
    drift: np.float64

    @staticmethod
    def schema() -> tb.Schema[IndividualBasicRow]:
        return tb.schema(
            IndividualBasicRow, order_columns=True, missing_columns="error", extra_columns="drop"
        )


# TODO: rename to Reinterpolated
@dataclass(frozen=True)
class IndividualReinterpolatedRow:
    """
    Describes the scalar data associated
    Returns:

    """

    #: Spectrum name without path and extension
    name: str

    #: Instrument name
    instrument: str

    #: Observation date/time in MJD
    mjd: np.float64

    #: Optional RV shift correction in km/s
    model: np.float64

    #: Median value of model (same for all spectra) in km/s
    rv_mean: np.float64

    #: Difference model - rv_mean in km/s
    rv_shift: np.float64

    SNR_5500: np.float64

    jdb: np.float64

    berv: np.float64

    lamp_offset: np.float64

    plx_mas: np.float64

    acc_sec: np.float64

    # same for all spectra
    hole_left: np.float64

    # same for all spectra
    hole_right: np.float64

    # same for all spectra
    wave_min: np.float64

    # same for all spectra
    wave_max: np.float64

    # same for all spectra
    dwave: np.float64

    # same for all spectra, len(pickled_spectrum.flux)
    nb_bins: np.int64

    #: Radial velocity
    vrad: np.float64

    #: Radial velocity uncertainty
    svrad: np.float64

    #: Instrumental drift
    drift: np.float64

    @staticmethod
    def schema() -> tb.Schema[IndividualReinterpolatedRow]:
        return tb.schema(
            IndividualReinterpolatedRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


class ReinterpolatedSpectrumPickle(TypedDict):
    """
    Data format of the pickle files produced by the reinterpolation step
    """

    #: Flux
    flux: NDArray[np.float64]
    #: Flux error
    flux_err: NDArray[np.float64]
    #: Average rv correction (median), same for all spectra
    RV_sys: np.float64
    #: RV correction, shift compared to the median
    RV_shift: np.float64
    #: Corresponds to the square root of the 95th percentile for 100 bins around the wavelength=5500
    SNR_5500: np.float64
    #: what is berv?
    berv: np.float64
    #: what is lamp offset?
    lamp_offset: np.float64
    #: what is plx_mas?
    plx_mas: np.float64
    #: what is acc_sec?
    acc_sec: np.float64
    #: instrument name
    instrument: str
    #: observation time in mjd
    mjd: np.float64
    #: what is jdb?
    jdb: np.float64
    #: Left boundary of hole, or -99.9 if not present
    hole_left: np.float64
    #: Right boundary of hole, or -99.9 if not present
    hole_right: np.float64
    #: Minimum wavelength
    wave_min: np.float64
    # TOCHECK: here
    #: Maximum wavelength, not necessarily equal to np.max(static_grid)
    wave_max: np.float64
    #: delta between two bins, synonym dlambda
    dwave: np.float64


class PickledIndividualSpectrum(TypedDict):
    """
    Data format of the pickle files produced by the preprocessing step
    """

    #: Wavelength in Angstroms
    wave: NDArray[np.float64]

    #: Flux in photon count units, must not have NaNs
    flux: NDArray[np.float64]

    #: Error on flux, must not have NaNs
    flux_err: NDArray[np.float64]

    #: Instrument name
    instrument: str

    #: Observation time in mjd
    mjd: np.float64

    #: Observation time in jdb
    jdb: np.float64

    #: Berv
    berv: np.float64

    #: Simultaneous drift in m/s
    lamp_offset: np.float64

    #: Parallax in milliarcseconds
    plx_mas: np.float64

    #: Secular acceleration
    acc_sec: np.float64

    #: np.min(self.wave)
    wave_min: np.float64

    #: np.max(self.wave)
    wave_max: np.float64

    #: Average delta between bins (note that dlambda is now set by a config parameter)
    dwave: np.float64


# TODO: DACE -> metatable
# TODO: remove the mjd patch

# TODO: rename to Imported
@dataclass(frozen=True)
class IndividualImportedRow:
    """Scalar values from individual pickles tabulated for ease of computation"""

    #: Spectrum name without path and extension
    name: str

    #: Instrument name
    instrument: str

    #: Observation date/time in MJD
    mjd: np.float64

    #: Optional RV shift correction in km/s
    model: np.float64

    #: Median value of model (same for all spectra) in km/s
    rv_mean: np.float64

    #: Difference model - rv_mean in km/s
    rv_shift: np.float64

    jdb: np.float64

    berv: np.float64

    #: Document
    #:
    #: Sometimes called lamp
    lamp_offset: np.float64

    plx_mas: np.float64

    acc_sec: np.float64

    wave_min: np.float64

    wave_max: np.float64

    dwave: np.float64

    hole_left: np.float64

    hole_right: np.float64

    #: Radial velocity
    vrad: np.float64

    #: Radial velocity uncertainty
    svrad: np.float64

    #: Instrumental drift
    drift: np.float64

    @staticmethod
    def schema() -> tb.Schema[IndividualImportedRow]:
        return tb.schema(
            IndividualImportedRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )



"""Parsed types used in RASSINE configuration"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Sequence, Tuple, Union

import configpile as cp
from typing_extensions import TypeAlias


@dataclass(frozen=True)
class RegPoly:
    """Penality-radius law, polynomial mapping, see Eq. (2) of RASSINE paper"""

    def name(self) -> str:
        return "poly"

    string: str
    expo: float


@dataclass(frozen=True)
class RegSigmoid:
    """Penality-radius law, logistic model"""

    def name(self) -> str:
        return "sigmoid"

    string: str
    center: float
    steepness: float


Reg: TypeAlias = Union[RegPoly, RegSigmoid]


@dataclass(frozen=True)
class Auto:
    ratio: float


Stretching: TypeAlias = Union[float, Auto]


def stretching_to_string(value: Stretching) -> str:
    if isinstance(value, Auto):
        return f"auto_{value.ratio}"
    else:
        return str(value)


def _stretching_parse(arg: str) -> Stretching:
    try:
        return float(arg)
    except:
        pass
    assert arg.startswith("auto_")
    return Auto(float(arg[5:]))


def _reg_parse(arg: str) -> Reg:
    parts = arg.split("_")
    if parts[0] == "poly":
        assert len(parts) == 2
        return RegPoly(string=arg, expo=float(parts[1]))
    elif parts[0] == "sigmoid":
        assert len(parts) == 3
        return RegSigmoid(string=arg, center=float(parts[1]), steepness=float(parts[2]))
    else:
        raise ValueError("Invalid reg")


reg_parser: cp.Parser[Reg] = cp.Parser.from_function_that_raises(_reg_parse)


stretching_parser: cp.Parser[Stretching] = cp.Parser.from_function_that_raises(_stretching_parse)


def _auto_float_parse(arg: str) -> Union[float, Literal["auto"]]:
    if arg.strip().lower() == "auto":
        return "auto"
    else:
        return float(arg)


auto_float_parser: cp.Parser[Union[float, Literal["auto"]]] = cp.Parser.from_function_that_raises(
    _auto_float_parse
)


def _auto_int_parse(arg: str) -> Union[int, Literal["auto"]]:
    if arg.strip().lower() == "auto":
        return "auto"
    else:
        return int(arg)


auto_int_parser: cp.Parser[Union[int, Literal["auto"]]] = cp.Parser.from_function_that_raises(
    _auto_int_parse
)


def _parse_mask_telluric(s: str) -> Sequence[Tuple[float, float]]:
    """
    Parses a list of telluric to mask

    Args:
        s: String argument to parse

    Returns:
        The parsed value
    """
    import parsy
    from parsy import seq, string

    # parse a number
    number = parsy.regex(r"-?(0|[1-9][0-9]*)([.][0-9]+)?([eE][+-]?[0-9]+)?").map(float)
    pair = (string("[") >> seq(number << string(","), number) << string("]")).map(
        lambda s: (s[0], s[1])
    )
    pairs = string("[") >> pair.sep_by(string(",")) << string("]")
    res: Sequence[Tuple[float, float]] = pairs.parse(s)
    return res


mask_telluric_parser = cp.Parser.from_function_that_raises(_parse_mask_telluric)



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

import configpile as cp
import pandas as pd
from typing_extensions import Annotated

from ..lib.data import LoggingLevel
from ..lib.util import log_task_name_and_time


@dataclass(frozen=True)
class Task(cp.Config):
    """
    Enumerates the row indices of a CSV table, one per row, into the standard output

    Specifically, this prints the integers 0 to n-1 included, where n is the number of rows
    in the CSV file.
    """

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #

    prog_ = Path(__file__).stem

    #: Input table to enumerate the row indices of
    input_table: Annotated[
        Path,
        cp.Param.store(
            cp.parsers.path_parser,
            long_flag_name=None,
            short_flag_name=None,
            positional=cp.Positional.ONCE,
        ),
    ]

    #: Column to get unique values from
    column: Annotated[str, cp.Param.store(cp.parsers.stripped_str_parser, short_flag_name="-c")]


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    table_path = t.root / t.input_table
    table = pd.read_csv(table_path)
    unique_values = table[t.column].unique()
    logging.debug(f"Got {len(unique_values)} unique values from {len(table)} rows of {table_path}")
    for v in unique_values:
        print(v)


def cli() -> None:
    run(Task.from_command_line_())



"""
Various dataclasses for standard RASSINE configuration parameters
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import configpile as cp
import tybles as tb

from . import io


@dataclass(frozen=True)
class NameRow:
    """
    Describes a table with a name column
    """

    #: Stacked spectrum name without path and extension
    name: str

    @staticmethod
    def schema() -> tb.Schema[NameRow]:
        return tb.schema(
            NameRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


@dataclass(frozen=True)
class PathPattern:
    """Path fragment, contains the relative path + filename pattern containing a {name} substring"""

    #: Relative path, possibly containing "{}"
    path: Path

    @staticmethod
    def parser() -> cp.Parser[PathPattern]:
        """Returns a configpile parser for a PathPattern"""
        return cp.parsers.path_parser.map(PathPattern)

    def to_path(self, root: Path, replacement: str) -> Path:
        """
        Returns a full path from this fragment

        Args:
            root: Absolute root path
            replacement: String to put in place of {name} in the fragment
        """
        assert root.is_absolute(), "Root path must be absolute"
        return root / Path(*[s.replace("{name}", replacement) for s in self.path.parts])


@dataclass(frozen=True)
class LoggingLevel:
    """Describes a logging level"""

    #: Logging level integer value, after parsing
    int_value: int

    @staticmethod
    def _parse(arg: str) -> cp.Res[LoggingLevel]:
        try:
            return LoggingLevel(int(arg))
        except ValueError:
            pass
        res = logging.getLevelName(arg)
        if isinstance(res, int):
            return LoggingLevel(res)
        return cp.Err.make(f"Cannot parse logging level {arg}")

    @classmethod
    def parser(cls) -> cp.Parser[LoggingLevel]:
        """Returns a parser for a logging level"""
        return cp.Parser.from_function(LoggingLevel._parse)

    def set(self, logger: Optional[logging.Logger] = None) -> None:
        """Sets the logging level in the provided logger

        Args:
            logger: Logger to update, or root logger (if :data:`None`)
        """
        if logger is None:
            logger = logging.getLogger()
        logger.setLevel(self.int_value)


@dataclass(frozen=True)
class PickleProtocol:
    """Describes a pickle protocol level"""

    level: int

    @classmethod
    def parser(cls) -> cp.Parser[PickleProtocol]:
        """Returns a parser for a Pickle protocol"""
        return cp.parsers.int_parser.map(PickleProtocol)

    def set(self) -> None:
        """Sets the pickle protocol level as parsed"""
        io.default_pickle_protocol = self.level



"""
Main file for the RASSINE script
"""

from __future__ import annotations

import argparse
import logging
import textwrap
import typing
from dataclasses import dataclass, replace
from datetime import datetime
from pathlib import Path
from typing import List, Literal, Mapping, Optional, Sequence, Tuple, Union

import configpile as cp
import matplotlib
import matplotlib.pylab as plt
import rich
import rich.pretty
import tybles as tb
from matplotlib.ticker import MultipleLocator
from typeguard import check_type
from typing_extensions import Annotated

from ..lib.data import LoggingLevel, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..stacking.types import MasterPickle, StackedBasicRow, StackedPickle
from .parsing import (
    Auto,
    Reg,
    Stretching,
    auto_float_parser,
    auto_int_parser,
    mask_telluric_parser,
    reg_parser,
    stretching_parser,
)
from .process import rassine_process
from .types import ExtraPlotData, RassineParameters, RassinePickle


@dataclass(frozen=True)
class Task(cp.Config):
    """
    Rolling Alpha Shape for a Spectral Improved Normalisation Estimator (RASSINE)

    ..
    
             ^                  .-=-.          .-==-.
            {}      __        .' O o '.       /   ^  )
           { }    .' O'.     / o .-. O \\     /  .--`\\
           { }   / .-. o\\   /O  /   \\  o\\   /O /    ^  (RASSSSSSINE)
            \\ `-` /   \\ O`-'o  /     \\  O`-`o /
        jgs  `-.-`     '.____.'       `.____.'

    Authors: Michael Cretignier, Jeremie Francfort and Denis Rosset
    """

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #

    prog_ = Path(__file__).stem

    ini_strict_sections_ = [Path(__file__).stem.split("_")[0]]

    #: Path of the spectrum pickle
    #:
    #: Either a path to a spectrum pickle needs to be provided, or one needs to provide
    #: the --input-table, --input-folder and inputs (positional) arguments
    #:
    #: Note that input_spectrum is a path relative to "root" and "input_folder"
    input_spectrum: Annotated[
        Optional[Path],
        cp.Param.store(
            cp.parsers.path_parser.empty_means_none(), short_flag_name="-s", default_value=""
        ),
    ]

    #: Table of stacked spectra from which to read the individual file names
    input_table: Annotated[
        Optional[Path],
        cp.Param.store(
            cp.parsers.path_parser.empty_means_none(), short_flag_name="-I", default_value=""
        ),
    ]

    #: Relative path to the folder containing the stacked spectra
    input_folder: Annotated[
        Optional[Path],
        cp.Param.store(
            cp.parsers.path_parser.empty_means_none(), short_flag_name="-i", default_value=""
        ),
    ]

    #: Indices of spectrum to process
    #:
    #: If provided, then input_table and input_folder must be provided
    #: If not provided, then input_spectrum must be provided
    input_indices: Annotated[
        Sequence[int],
        cp.Param.append1(
            cp.parsers.int_parser,
            positional=cp.Positional.ZERO_OR_MORE,
            long_flag_name=None,
            short_flag_name=None,
        ),
    ]

    def validate_inputs(self) -> Optional[cp.Err]:
        errors: List[Optional[cp.Err]] = []
        if self.input_indices:
            errors.append(
                cp.Err.check(
                    self.input_table is not None,
                    "If input_indices are provided, input_table must be provided",
                )
            )
            errors.append(
                cp.Err.check(
                    self.input_folder is not None,
                    "If input_indices are provided, input_folder must be provided",
                )
            )
            errors.append(
                cp.Err.check(
                    self.input_spectrum is None,
                    "If input_indices are provided, then input_spectrum must not be provided",
                )
            )
        else:
            errors.append(
                cp.Err.check(
                    self.input_spectrum is not None,
                    "At least one of input_indices or input_spectrum must be provided",
                )
            )
        if self.input_spectrum is not None:
            errors.append(
                cp.Err.check(
                    self.input_table is None,
                    "If input_spectrum is provided, then input_table must not be provided",
                )
            )
        return cp.Err.collect(*errors)

    #: Directory where output files are written
    output_folder: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, short_flag_name="-o"),
    ]

    #: Pattern to use for the RASSINE output filename
    #:
    #: The "{}" part will be replaced by the input filename stem
    output_pattern: Annotated[
        str, cp.Param.store(cp.parsers.str_parser, default_value="RASSINE_{}.p")
    ]

    #: Put a RASSINE output file that will fix the value of the 7 parameters to the same value than in the anchor file
    input_anchor_pickle: Annotated[
        Optional[Path],
        cp.Param.store(
            cp.parsers.path_parser.empty_means_none(), short_flag_name="-l", default_value=""
        ),
    ]

    def validate_input_anchor_pickle(self) -> Optional[cp.Err]:
        return cp.Err.check(
            self.input_anchor_pickle is None or (self.root / self.input_anchor_pickle).exists(),
            "Anchor pickle file, if provided, must exist",
        )

    #: Optional INI output anchor file that can be used as an input configuration file
    #:
    #: Only valid if input_spectrum is used and not the multi-spectrum reduction
    output_anchor_ini: Annotated[
        Optional[Path],
        cp.Param.store(cp.parsers.path_parser.empty_means_none(), default_value=""),
    ]

    def validate_output_anchor_ini(self) -> Optional[cp.Err]:
        if self.output_anchor_ini is not None:
            return cp.Err.check(
                self.input_spectrum is not None,
                "If output_anchor_ini is provided, then one must provide input_spectrum as well",
            )
        return None

    #: Where to put the output plot, if empty no plot is produced
    #:
    #: The plot name will be derived from the spectrum name, with a "_output.png" suffix
    output_plot_folder: Annotated[
        Optional[Path],
        cp.Param.store(cp.parsers.path_parser.empty_means_none(), default_value=""),
    ]

    #: Pattern to use for the RASSINE output filename
    #:
    #: The "{}" part will be replaced by the input filename stem
    output_plot_pattern: Annotated[
        str,
        cp.Param.store(cp.parsers.str_parser.empty_means_none(), default_value="{}_output.png"),
    ]

    #: stretch the x and y axes ratio ('auto' available)
    #:
    #: PARAMETER 1
    par_stretching: Annotated[
        Stretching,
        cp.Param.store(stretching_parser, default_value="auto_0.5", short_flag_name="-p"),
    ]

    def validate_par_stretching(self) -> Optional[cp.Err]:
        if isinstance(self.par_stretching, Auto):
            if not (0 <= self.par_stretching.ratio <= 1):
                return cp.Err.make("the par_stretching auto value must be between 0 and 1")
        else:
            if self.par_stretching < 0.0:
                return cp.Err.make("the par_stretching fixed value cannot be negative")
        return None

    #: half-window to find a local maxima
    par_vicinity: Annotated[int, cp.Param.store(cp.parsers.int_parser, default_value="7")]

    #: half-window of the box used to smooth (1 => no smoothing, 'auto' available)
    #:
    #: PARAMETER 2
    par_smoothing_box: Annotated[
        Union[Literal["auto"], int], cp.Param.store(auto_int_parser, default_value="6")
    ]
    #: Smoothing kernel, can take the values:
    #:
    #: 'rectangular','gaussian','savgol' if a value is specified in smoothing box
    #: 'erf','hat_exp' if 'auto' is given in smoothing box
    par_smoothing_kernel: Annotated[
        Literal["rectangular", "gaussian", "savgol", "erf", "hat_exp"],
        cp.Param.store(
            cp.Parser.from_choices(["rectangular", "gaussian", "savgol", "erf", "hat_exp"]),
            default_value="savgol",
        ),
    ]

    def validate_smoothing(self) -> Optional[cp.Err]:
        if self.par_smoothing_box == "auto":
            if self.par_smoothing_kernel not in ["erf", "hat_exp"]:
                return cp.Err.make(
                    "If par_smoothing_box is auto, par_smoothing_kernel should be erf or hat_exp"
                )
        else:
            if self.par_smoothing_kernel not in ["rectangular", "gaussian", "savgol"]:
                return cp.Err.make(
                    "If par_smoothing_box is an integer, par_smoothing_kernel should be rectangular, gaussian or savgol"
                )
        return None

    #: FWHM of the CCF in km/s ('auto' available)
    #:
    #: PARAMETER 3
    par_fwhm: Annotated[
        Union[Literal["auto"], float], cp.Param.store(auto_float_parser, default_value="auto")
    ]

    #: only needed if par_fwhm is in 'auto'
    CCF_mask: Annotated[str, cp.Param.store(cp.parsers.str_parser, default_value="master")]

    #: Minimum radius of the rolling pin in angstrom ('auto' available)
    #:
    #: PARAMETER 4
    par_R: Annotated[
        Union[Literal["auto"], float],
        cp.Param.store(auto_float_parser, default_value="auto", short_flag_name="-r"),
    ]

    #: Maximum radius of the rolling pin in angstrom ('auto' available)
    #:
    #: PARAMETER 5
    par_Rmax: Annotated[
        Union[Literal["auto"], float],
        cp.Param.store(auto_float_parser, default_value="auto", short_flag_name="-R"),
    ]

    #: Penality-radius law
    #:
    #: poly_d (d the degree of the polynome x**d)
    #: or sigmoid_c_s where c is the center and s the steepness
    #:
    #: PARAMETER 6
    par_reg_nu: Annotated[Reg, cp.Param.store(reg_parser, default_value="poly_1.0")]

    #: a list of left and right borders to eliminate from the mask of the CCF
    #: only if CCF = 'master' and par_fwhm = 'auto'
    mask_telluric: Annotated[
        Sequence[Tuple[float, float]],
        cp.Param.store(
            mask_telluric_parser, default_value="[[6275,6330],[6470,6577],[6866,8000]]"
        ),
    ]

    #: True if working with a noisy-free synthetic spectra
    synthetic_spectrum: Annotated[
        bool, cp.Param.store(cp.parsers.bool_parser, default_value="false")
    ]

    #: Define the interpolation for the continuum displayed in the subproducts
    #:
    #: note that at the end a cubic and linear interpolation are saved in 'output' regardless this value
    interpolation: Annotated[
        Literal["cubic", "linear"],
        cp.Param.store(cp.Parser.from_choices(["cubic", "literal"]), default_value="cubic"),
    ]

    #: Half window of the area used to average the number of point around the local max for the continuum
    denoising_dist: Annotated[int, cp.Param.store(auto_int_parser, default_value="5")]

    #: Number of border cut in automatic mode (put at least 3 if Automatic mode)
    count_cut_lim: Annotated[int, cp.Param.store(auto_int_parser, default_value="3")]

    #: Number of outliers clipping in automatic mode (put at least 1 if Automatic mode)
    count_out_lim: Annotated[int, cp.Param.store(auto_int_parser, default_value="1")]

    #: Random seed, derived from spectrum name if empty
    random_seed: Annotated[
        Optional[int], cp.Param.store(cp.parsers.int_parser.empty_means_none(), default_value="")
    ]


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


def print_parameters_according_to_paper(parameters: RassineParameters):
    translation: Mapping[str, str] = {
        "filename": "filename",
        "number_iteration": "number_iteration",
        "K_factors": "K_factors",
        "axes_stretching": "par_stretching",
        "vicinity_local_max": "par_vicinity",
        "smoothing_box": "par_smoothing_box",
        "smoothing_kernel": "par_smoothing_kernel",
        "fwhm_ccf": "par_fwhm",
        "CCF_mask": "CCF_mask",
        "RV_sys": "RV_sys",
        "min_radius": "par_R",
        "max_radius": "par_Rmax",
        "model_penality_radius": "par_reg_nu",
        "denoising_dist": "denoising_dist",
        "number_of_cut": "count_cut_lim",
        "windows_penality": "windows_penality",
        "large_window_penality": "large_window_penality",
        "number_points": "number of points",
        "number_anchors": "number of anchors",
        "SNR_5500": "SNR_5500",
        "mjd": "mjd",
        "jdb": "jdb",
        "wave_min": "wave_min",
        "wave_max": "wave_max",
        "dwave": "dwave",
        "hole_left": "hole_left",
        "hole_right": "hole_right",
        "RV_shift": "RV_shift",
        "berv": "berv",
        "lamp_offset": "lamp_offset",
        "acc_sec": "acc_sec",
        "light_file": "light_file",
        "speedup": "speedup",
        "continuum_interpolated_saved": "continuum_interpolated_saved",
        "continuum_denoised_saved": "continuum_denoised_saved",
        "nb_spectra_stacked": "nb_spectra_stacked",
        "arcfiles": "arcfiles",
        "rv_mean_jdb": "rv_mean_jdb",
        "rv_dace": "rv_dace",
        "rv_dace_std": "rv_dace_std",
    }
    rich.print("[bold red]Output parameters[/bold red]")
    new_dict = {translation[key]: value for key, value in parameters.items()}
    rich.pretty.pprint(new_dict)


def plot_output(
    output: RassinePickle,
    extra: ExtraPlotData,
    output_file: Optional[Path] = None,
    jump_point: int = 1,
):
    """Plots RASSINE result

    Args:
        jump_point: Put > 1 to make lighter plot for article
    """
    grid = output["wave"]
    spectrei = output["flux"]
    wave = output["output"]["anchor_wave"]
    flux = output["output"]["anchor_flux"]
    normalisation = extra.normalisation
    spectre = extra.spectre
    conti = extra.conti
    SNR_0 = output["parameters"]["SNR_5500"]  # now rounded to integer
    fig = plt.figure(figsize=(16, 6))
    plt.subplot(2, 1, 1)
    plt.plot(
        grid[::jump_point],
        spectrei[::jump_point],
        label=f"spectrum (SNR={int(SNR_0):.0f})",
        color="g",
    )
    plt.plot(
        grid[::jump_point],
        spectre[::jump_point] * normalisation,
        label="spectrum reduced",
        color="b",
        alpha=0.3,
    )
    plt.scatter(wave, flux, color="k", label=f"anchor points ({int(len(wave))})", zorder=100)
    plt.plot(grid[::jump_point], conti[::jump_point], label="continuum", zorder=101, color="r")
    plt.xlabel("Wavelength", fontsize=14)
    plt.ylabel("Flux", fontsize=14)
    plt.legend(loc=4)
    plt.title("Final products of RASSINE", fontsize=14)
    ax = plt.gca()
    ax.xaxis.set_minor_locator(MultipleLocator(50))
    plt.tick_params(direction="in", top=True, which="both")
    plt.subplot(2, 1, 2, sharex=ax)
    plt.plot(grid[::jump_point], spectrei[::jump_point] / conti[::jump_point], color="k")
    plt.axhline(y=1, color="r", zorder=102)
    plt.xlabel(r"Wavelength [$\AA$]", fontsize=14)
    plt.ylabel("Flux normalised", fontsize=14)
    ax = plt.gca()
    ax.xaxis.set_minor_locator(MultipleLocator(50))
    plt.tick_params(direction="in", top=True, which="both")
    plt.subplots_adjust(left=0.07, right=0.96, hspace=0, top=0.95)
    if output_file is not None:
        plt.savefig(output_file)
    # TODO: remove this
    plt.close()


def update_using_anchor_file(task: Task, anchor_file: Path) -> Task:
    # TODO: update this
    data = open_pickle(anchor_file)
    return replace(
        task,
        par_stretching=data["parameters"]["axes_stretching"],
        par_vicinity=data["parameters"]["vicinity_local_max"],
        par_smoothing_box=data["parameters"]["smoothing_box"],
        par_smoothing_kernel=data["parameters"]["smoothing_kernel"],
        par_fwhm=data["parameters"]["fwhm_ccf"],
        par_R=data["parameters"]["min_radius"],
        par_Rmax=data["parameters"]["max_radius"],
        par_reg_nu=data["parameters"]["model_penality_radius"],
        count_cut_lim=data["parameters"]["number_of_cut"],
    )


def make_ini_contents_from_parameters(
    params: RassineParameters, input_file: Optional[Path] = None
) -> str:
    return textwrap.dedent(
        f"""
        # Anchor file created on {datetime.now().isoformat()} from {input_file}
        [rassine]
        par-stretching={params["axes_stretching"]}
        par-vicinity={params["vicinity_local_max"]}
        par-smoothing-box={params["smoothing_box"]}
        par-smoothing-kernel={params["smoothing_kernel"]}
        par-fwhm={params["fwhm_ccf"]}
        par-R={params["min_radius"]}
        par-Rmax={params["max_radius"]}
        par-reg-nu={params["model_penality_radius"]}
        count-cut-lim={params["number_of_cut"]}
    """
    ).strip()


def run(t: Task):
    t.pickle_protocol.set()
    t.logging_level.set()
    if logging.getLogger().level <= 20:  # logging INFO
        rich.pretty.pprint(t)

    if t.input_anchor_pickle is not None:
        update_using_anchor_file(t, t.root / t.input_anchor_pickle)

    (t.root / t.output_folder).mkdir(parents=True, exist_ok=True)

    if t.output_anchor_ini is not None:
        (t.root / t.output_anchor_ini).parent.mkdir(parents=True, exist_ok=True)

    if t.output_plot_folder is not None:
        (t.root / t.output_plot_folder).mkdir(parents=True, exist_ok=True)

    @dataclass(frozen=True)
    class Step:
        stem: str
        row: Optional[StackedBasicRow]
        input_file: Path
        output_file: Path
        output_plot_file: Optional[Path]

    def output_plot_file_for_stem(stem: str) -> Optional[Path]:
        if t.output_plot_folder is not None:
            return t.root / t.output_plot_folder / (t.output_plot_pattern.replace("{}", stem))
        else:
            return None

    if t.input_indices:
        # do batch processing
        assert t.input_table is not None
        assert t.input_folder is not None
        logging.debug(f"Reading {t.root/t.input_table}")
        tyble = StackedBasicRow.schema().read_csv(t.root / t.input_table, return_type="Tyble")
        stems = [tyble[i].name for i in t.input_indices]
        steps: List[Step] = []
        for i in t.input_indices:
            row = tyble[i]
            stem = row.name
            steps.append(
                Step(
                    stem=stem,
                    row=tyble[i],
                    input_file=t.root / t.input_folder / (stem + ".p"),
                    output_file=t.root / t.output_folder / (t.output_pattern.replace("{}", stem)),
                    output_plot_file=output_plot_file_for_stem(stem),
                )
            )
    else:
        if t.input_folder is None:
            abs_input_folder = t.root
        else:
            abs_input_folder = t.root / t.input_folder
        assert t.input_spectrum is not None
        stem = t.input_spectrum.stem
        steps = [
            Step(
                stem,
                None,
                abs_input_folder / t.input_spectrum,
                t.root / t.output_folder / (t.output_pattern.replace("{}", stem)),
                output_plot_file_for_stem(stem),
            )
        ]

    if t.output_plot_folder is not None:
        matplotlib.use("Qt5Agg", force=True)
        plt.close("all")

    for step in steps:
        # compute random seed from input filename if not provided, so the randomization
        # is deterministic between runs on the same files
        if t.random_seed is None:
            random_seed = hash(step.input_file.stem)
        else:
            random_seed = t.random_seed

        data: Union[MasterPickle, StackedPickle] = open_pickle(step.input_file)

        data_id: List[str] = []
        try:
            check_type("data", data, MasterPickle)
            data_id.append("MasterPickle")
        except TypeError as e:
            pass

        try:
            check_type("data", data, StackedPickle)
            data_id.append("StackedPickle")
        except TypeError as e:
            pass

        logging.info(f"Read input pickle {step.input_file}, types identified: {data_id}")

        output, extra_plot_data = rassine_process(
            output_filename=step.output_file.name,
            row=step.row,
            data=data,
            synthetic_spectrum=t.synthetic_spectrum,
            random_seed=random_seed,
            par_stretching=t.par_stretching,
            par_vicinity=t.par_vicinity,
            par_smoothing_box=t.par_smoothing_box,
            par_smoothing_kernel=t.par_smoothing_kernel,
            par_fwhm_=t.par_fwhm,
            CCF_mask=t.CCF_mask,
            mask_telluric=t.mask_telluric,
            par_R_=t.par_R,
            par_Rmax_=t.par_Rmax,
            par_reg_nu=t.par_reg_nu,
            denoising_dist=t.denoising_dist,
            count_cut_lim=t.count_cut_lim,
            count_out_lim=t.count_out_lim,
            interpol=t.interpolation,
        )
        params = output["parameters"]
        save_pickle(step.output_file, output)
        print(
            f"Output file saved under : {step.output_file} (SNR at 5500 : {params['SNR_5500']:.0f})"
        )

        if logging.getLogger().level <= 20:  # logging INFO
            print_parameters_according_to_paper(output["parameters"])

        if t.output_anchor_ini is not None:
            with open(t.root / t.output_anchor_ini, "wt") as f:
                f.write(make_ini_contents_from_parameters(output["parameters"], step.input_file))

        if step.output_plot_file is not None:
            plot_output(output, extra_plot_data, step.output_plot_file)


def cli() -> None:
    run(Task.from_command_line_())



from __future__ import annotations

import typing  # pylint: disable=W0611
from dataclasses import dataclass
from typing import Literal, Sequence, TypedDict

import numpy as np
import tybles as tb
from numpy.typing import NDArray
from typing_extensions import Annotated


@dataclass(frozen=True)
class StackedBasicRow:
    """
    Describes the scalar data associated with a stacked spectrum
    """

    #: Stacked spectrum name without path and extension
    name: str
    #: Group index
    group: int
    #: Instrument name
    instrument: str
    #: mjd weighted average
    mjd: np.float64
    #: Average rv correction (median), same for all spectra
    rv_mean: np.float64
    #: RV correction, shift compared to the median, weighted average
    rv_shift: np.float64
    #: Corresponds to the square root of the 95th percentile for 100 bins around the wavelength=5500
    SNR_5500: np.float64
    #: jdb weighted average
    jdb: np.float64
    #: berv, weighted average
    berv: np.float64
    #: min(berv) for the individual spectra in the group
    berv_min: np.float64
    #: max(berv) for the individual spectra in the group
    berv_max: np.float64
    #: lamp_offset, weighted average
    lamp_offset: np.float64
    #: acc_sec, taken from first spectrum
    acc_sec: np.float64
    #: Left boundary of hole, or -99.9 if not present
    hole_left: np.float64
    #: Right boundary of hole, or -99.9 if not present
    hole_right: np.float64
    #: Minimum wavelength
    wave_min: np.float64
    #: Maximum wavelength, not necessarily equal to np.max(static_grid)
    wave_max: np.float64
    #: delta between two bins, synonym dlambda
    dwave: np.float64
    #: Number of days using for the stacking
    stacking_length: np.float64
    #: Number of individual spectra using for this individual spectrum
    nb_spectra_stacked: np.int64
    # same for all spectra, len(flux)
    nb_bins: np.int64

    #: Time weighted by the radial velocity uncertainty (info for YARARA)
    mean_jdb: np.float64

    #: Radial velocity weighted by the radial velocity uncertainty (info for YARARA)
    #:
    #: Median *not* recentered over all averages, YARAR takes care of that
    mean_vrad: np.float64

    #: Propagated harmonic mean of the radial velocity uncertainties (info for YARARA)
    mean_svrad: np.float64

    @staticmethod
    def schema() -> tb.Schema[StackedBasicRow]:
        return tb.schema(
            StackedBasicRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


class StackedPickle(TypedDict):
    """
    Data format of the pickle files produced by the stacking step

    All the weighted averages are made using the bolometric flux
    """

    #: Flux, stacked
    flux: NDArray[np.float64]
    #: Flux error, stacked (square of this is the sum of squares of individual spectra)
    flux_err: NDArray[np.float64]
    #: Average rv correction (median), same for all spectra
    RV_sys: np.float64
    #: RV correction, shift compared to the median, weighted average
    RV_shift: np.float64
    #: Corresponds to the square root of the 95th percentile for 100 bins around the wavelength=5500
    SNR_5500: np.float64
    #: berv, weighted average
    berv: np.float64
    #: lamp_offset, weighted average
    lamp_offset: np.float64
    #: acc_sec, taken from first spectrum
    acc_sec: np.float64
    #: mjd weighted average
    mjd: np.float64
    #: jdb weighted average
    jdb: np.float64
    #: Left boundary of hole, or -99.9 if not present
    hole_left: np.float64
    #: Right boundary of hole, or -99.9 if not present
    hole_right: np.float64
    #: Minimum wavelength
    wave_min: np.float64
    #: Maximum wavelength, not necessarily equal to np.max(static_grid)
    wave_max: np.float64
    #: delta between two bins, synonym dlambda
    dwave: np.float64
    #: Number of days using for the stacking
    stacking_length: np.float64
    #: Number of individual spectra using for this individual spectrum
    nb_spectra_stacked: int
    #: Paths of files used in this stacked spectrum
    arcfiles: Sequence[str]


@dataclass(frozen=True)
class MasterRow:
    name: str
    SNR_5500: np.float64
    acc_sec: np.float64
    berv: np.float64
    berv_min: np.float64
    berv_max: np.float64
    instrument: str
    hole_left: np.float64
    hole_right: np.float64
    wave_min: np.float64
    wave_max: np.float64
    dwave: np.float64
    nb_spectra_stacked: int

    @staticmethod
    def schema() -> tb.Schema[MasterRow]:
        return tb.schema(
            MasterRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


@dataclass(frozen=True)
class IndividualGroupRow:
    """Columns of the DACE data we extract for the rest of the processing"""

    #: Spectrum name without path and extension
    name: str

    #: Group index
    group: int

    #: Number of days used for stacking
    stacking_length: np.float64

    #: dbin to shift the binning (0.5 for solar data)
    dbin: np.float64

    @staticmethod
    def schema() -> tb.Schema[IndividualGroupRow]:
        return tb.schema(
            IndividualGroupRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


class MasterPickle(TypedDict):
    """
    Data format of the pickle files produced by the stacking step

    All the weighted averages are made using the bolometric flux
    """

    #: Flux, stacked
    flux: NDArray[np.float64]
    #: Tells it is the master spectrum
    master_spectrum: Literal[True]
    #: Average rv correction (median), same for all spectra
    RV_sys: np.float64
    #: RV correction, shift compared to the median, weighted average
    RV_shift: Annotated[np.float64, np.float64(0.0)]
    #: Corresponds to the square root of the 95th percentile for 100 bins around the wavelength=5500
    SNR_5500: np.float64
    #: lamp_offset, weighted average
    lamp_offset: Annotated[np.float64, np.float64(0.0)]
    #: acc_sec, taken from first spectrum
    acc_sec: np.float64
    #: berv, weighted average according to SNR
    berv: np.float64
    #: np.min(berv) of the individual spectra
    berv_min: np.float64
    #: np.max(berv) of the individual spectra
    berv_max: np.float64
    #: Instrument
    instrument: str
    #: mjd weighted average
    mjd: Annotated[np.float64, np.float64(0.0)]
    #: jdb weighted average
    jdb: Annotated[np.float64, np.float64(0.0)]
    #: Left boundary of hole, or -99.9 if not present
    hole_left: np.float64
    #: Right boundary of hole, or -99.9 if not present
    hole_right: np.float64
    #: Minimum wavelength
    wave_min: np.float64
    #: Maximum wavelength, not necessarily equal to np.max(static_grid)
    wave_max: np.float64
    #: delta between two bins, synonym dlambda
    dwave: np.float64
    #: Number of individual spectra using for this individual spectrum
    nb_spectra_stacked: int
    #: Paths of files used in this stacked spectrum
    arcfiles: Literal["none"]



from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ..lib.math import Float, doppler_r


def empty_ccd_gap(
    wave: NDArray[np.float64],
    flux: NDArray[np.float64],
    left: Optional[Float] = None,
    right: Optional[Float] = None,
    extended: float = 30.0,
) -> NDArray[np.float64]:
    """
    Ensure a 0 value in the gap between the ccd of HARPS s1d with extended=30 kms extension

    Args:
        wave: Wavelength vector.
        flux: Flux vector.
        left: Wavelength of the left CCF gap.
        right: Wavelength of the right CCF gap.
        extended: Extension of the gap in kms.

    Returns:
        Flux values with null values inside the specified gap
    """

    dgrid = np.diff(wave)[0]

    if left is not None:
        left_ = np.array([left]).astype("float64")
        left = doppler_r(left_, -extended)[0][0]  # 30 km/s supression
    else:
        left = wave.max()

    if right is not None:
        right_ = np.array([right]).astype("float64")
        right = doppler_r(right_, extended)[0][0]  # 30 km/s supression
    else:
        right = wave.min()

    flux[(wave >= left - dgrid / 2) & (wave <= right + dgrid / 2)] = 0
    return flux



import pickle
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

import configpile as cp
import rich.pretty
from recursive_diff import recursive_diff
from typing_extensions import Annotated


@dataclass(frozen=True)
class Task(cp.Config):
    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    prog_ = Path(__file__).stem

    #: First pickle to compare
    input1: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, positional=cp.Positional.ONCE, long_flag_name=None),
    ]

    #: Second pickle to compare
    input2: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, positional=cp.Positional.ONCE, long_flag_name=None),
    ]

    #: Subdico to inspect, several keys may be separated by dots
    key: Annotated[str, cp.Param.store(cp.parsers.stripped_str_parser, default_value="")]


def run(t: Task):
    with open(t.root / t.input1, "rb") as file:
        data1 = pickle.load(file)
    with open(t.root / t.input2, "rb") as file:
        data2 = pickle.load(file)
    subkeys = [s.strip() for s in t.key.split(".") if s.strip() != ""]
    for key in subkeys:
        data1 = data1[key]
        data2 = data2[key]
    rich.pretty.pprint(list(recursive_diff(data1, data2)))


def cli():
    run(Task.from_command_line_())



import logging
import pickle
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Type, TypeVar, Union, overload

import numpy.typing as npt
import pandas as pd
from astropy.io import fits

default_pickle_protocol: int = 3

from typeguard import check_type


class NoCheck(Enum):
    TOKEN = 0


_T = TypeVar("_T")


@overload
def open_pickle(filename: Path, expected_type: Type[_T]) -> _T:
    pass


@overload
def open_pickle(filename: Path, expected_type: NoCheck = NoCheck.TOKEN) -> Any:
    pass


def open_pickle(filename: Path, expected_type: Any = NoCheck.TOKEN) -> Any:
    """
    Load information from a pickle

    Args:
        filename: Pickle to read
        expected_type: Optional type to validate the data read from

    Raises:
        TypeError: if there is a type mismatch

    Returns:
        Information read
    """
    logging.debug("Reading pickle %s", filename)
    res = pd.read_pickle(filename)
    if not isinstance(expected_type, NoCheck):
        check_type("pickle", res, expected_type)
    return res


def save_pickle(filename: Path, output: Any, protocol: int = default_pickle_protocol):
    """
    Save a pickle file with the proper protocol pickle version.

    Args:
        filename: Name of the output pickle file.
        output: Data to save
        protocol:
    """
    logging.debug("Writing pickle %s", filename)
    with open(filename, "wb") as f:
        pickle.dump(output, f, protocol=protocol)



from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import List, Literal, Sequence, Tuple, TypedDict

import configpile as cp
import numpy as np
import tybles as tb
from astropy.io import fits
from filelock import FileLock
from numpy.typing import NDArray
from typing_extensions import Annotated

from ..lib.analysis import grouping
from ..lib.data import LoggingLevel, PathPattern, PickleProtocol
from ..lib.io import save_pickle
from ..lib.math import absurd_minus_99_9
from ..lib.util import log_task_name_and_time
from .types import IndividualBasicRow, IndividualImportedRow, PickledIndividualSpectrum


@dataclass(frozen=True)
class Task(cp.Config):
    """Import FITS files into pickle files that can be processed by RASSINE"""

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #

    prog_ = Path(__file__).stem

    ini_strict_sections_ = [Path(__file__).stem.split("_")[0]]

    #: Input spectrum table
    input_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-I")]

    #: Output spectrum table
    output_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-O")]

    #: Relative path to the folder containing the raw spectra
    input_folder: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-i")]

    #: Path pattern for output files
    output_pattern: Annotated[
        PathPattern, cp.Param.store(PathPattern.parser(), short_flag_name="-o")
    ]

    #: Indices of spectrum to process
    #:
    #: If not provided, all spectra are processed
    inputs: Annotated[
        Sequence[int],
        cp.Param.append1(
            cp.parsers.int_parser,
            positional=cp.Positional.ZERO_OR_MORE,
            long_flag_name=None,
            short_flag_name=None,
        ),
    ]

    #: Instrument format of the s1d spectra
    instrument: Annotated[
        Literal["HARPS", "CORALIE", "HARPN", "ESPRESSO", "EXPRESS"],
        cp.Param.store(
            cp.Parser.from_choices(["HARPS", "CORALIE", "HARPN", "ESPRESSO", "EXPRESS"]),
            default_value="HARPS",
        ),
    ]

    #: Parallax in mas (no more necessary ?)
    plx_mas: Annotated[
        float,
        cp.Param.store(cp.parsers.float_parser, default_value="0.0"),
    ]

    #: Type of DRS format
    drs_style: Annotated[
        Literal["old", "new"], cp.Param.store(cp.Parser.from_choices(["old", "new"]))
    ]


def find_hole(wave: NDArray[np.float64], flux: NDArray[np.float64]) -> Tuple[float, float]:
    """
    Finds a gap between CCD

    Args:
        wave: Wavelength array
        flux: Flux array

    Returns:
        The endpoints of the wavelength interval if the gap, or (-99.9, -99.9) if not found
    """
    null_flux = np.where(flux == 0)[0]  # criterion to detect gap between ccd
    left = absurd_minus_99_9
    right = absurd_minus_99_9
    if len(null_flux) > 0:
        mask = grouping(np.diff(null_flux), 0.5, 0)[-1]
        highest = mask[mask[:, 2].argmax()]
        if highest[2] > 1000:
            left = wave[int(null_flux[highest[0]])]  # store left or -99
            right = wave[int(null_flux[highest[1]])]  # store right or -99
    return (left, right)


def preprocess_import(
    row: IndividualBasicRow,
    header: fits.Header,
    data: np.ndarray,
    instrument: Literal["HARPS", "CORALIE", "HARPN", "ESPRESSO", "EXPRESS"],
    plx_mas: float,
    drs_style: Literal["old", "new"],
) -> Tuple[PickledIndividualSpectrum, IndividualImportedRow]:

    if drs_style == "old":
        spectre = data.astype("float64")  # the flux of your spectrum
        spectre_step = np.round(header["CDELT1"], 8)
        spectre_error = np.zeros(len(spectre))
        wave_min = np.round(header["CRVAL1"], 8)  # to round float32
        wave_max = np.round(
            header["CRVAL1"] + (len(spectre) - 1) * spectre_step, 8
        )  # to round float32
        wave = np.round(np.linspace(wave_min, wave_max, len(spectre)), 8)

        # cut left and right parts with zero flux
        # and reevaluate wave_min and wave_max

        # should this be done outside?
        begin = np.min(np.arange(len(spectre))[spectre > 0])
        end = np.max(np.arange(len(spectre))[spectre > 0])
        wave = wave[begin : end + 1]
        spectre = spectre[begin : end + 1]
        kw = "ESO"
        if instrument == "HARPN":
            kw = "TNG"
        berv = np.float64(header[f"HIERARCH {kw} DRS BERV"])
        lamp = np.float64(header[f"HIERARCH {kw} DRS CAL TH LAMP OFFSET"])
    else:
        spectre = data["flux"].astype("float64")  # the flux of your spectrum
        spectre_error = data["error"].astype("float64")  # the flux of your spectrum
        # wave was grid in the previous code
        wave = data["wavelength_air"].astype(
            "float64"
        )  # the grid of wavelength of your spectrum (assumed equidistant in lambda)
        begin = np.min(np.arange(len(spectre))[spectre > 0])  # remove border spectrum with 0 value
        end = np.max(np.arange(len(spectre))[spectre > 0])  # remove border spectrum with 0 value
        wave = wave[begin : end + 1]
        spectre = spectre[begin : end + 1]
        spectre_error = spectre_error[begin : end + 1]
        spectre_step = np.mean(np.diff(wave))
        kw = "ESO"
        if "HIERARCH TNG QC BERV" in header:
            kw = "TNG"
        berv = np.float64(header["HIERARCH " + kw + " QC BERV"])
        lamp = np.float64(0.0)  # header['HIERARCH ESO DRS CAL TH LAMP OFFSET'] no yet available

    wave_min = np.min(wave)
    wave_max = np.max(wave)

    if instrument == "CORALIE":
        if np.mean(spectre) < 100000:
            spectre *= 400780143771.18976  # calibrated with HD8651 2016-12-16 AND 2013-10-24

        spectre /= 1.4e10 / 125**2  # calibrated to match with HARPS SNR

    mjd = row.mjd

    pma = np.float64(0.0)
    pmd = np.float64(0.0)
    if f"HIERARCH {kw} TEL TARG PMA" in header:
        assert f"HIERARCH {kw} TEL TARG PMD" in header
        pma = np.float64(header[f"HIERARCH {kw} TEL TARG PMA"] * 1000)
        pmd = np.float64(header[f"HIERARCH {kw} TEL TARG PMD"] * 1000)

    if plx_mas:
        distance_m = 1000.0 / plx_mas * 3.08567758e16
        mu_radps = (
            np.sqrt(pma**2 + pmd**2) * 2 * np.pi / (360.0 * 1000.0 * 3600.0 * 86400.0 * 365.25)
        )
        acc_sec = distance_m * 86400.0 * mu_radps**2  # rv secular drift in m/s per days
    else:
        acc_sec = 0.0
    jdb = np.float64(mjd) + 0.5

    hole_left, hole_right = find_hole(wave, spectre)
    if hole_left != absurd_minus_99_9 and hole_right != absurd_minus_99_9:
        logging.info(f"Gap detected in s1d between {hole_left:.2f} and {hole_right:.2f}")

    output_pickle: PickledIndividualSpectrum = {
        "wave": wave,
        "flux": spectre,
        "flux_err": spectre_error,
        "instrument": instrument,
        "mjd": mjd,
        "jdb": np.float64(jdb),
        "berv": np.float64(berv),
        "lamp_offset": np.float64(lamp),
        "plx_mas": np.float64(plx_mas),
        "acc_sec": np.float64(acc_sec),
        "wave_min": wave_min,
        "wave_max": wave_max,
        "dwave": spectre_step,
    }

    output_row = IndividualImportedRow(
        name=row.name,
        instrument=instrument,
        mjd=mjd,
        model=row.model,
        rv_mean=row.rv_mean,
        rv_shift=row.rv_shift,
        jdb=jdb,
        berv=berv,
        lamp_offset=lamp,
        plx_mas=np.float64(plx_mas),
        acc_sec=np.float64(acc_sec),
        wave_min=wave_min,
        wave_max=wave_max,
        dwave=spectre_step,
        hole_left=np.float64(hole_left),
        hole_right=np.float64(hole_right),
        vrad=row.vrad,
        svrad=row.svrad,
        drift=row.drift,
    )
    return output_pickle, output_row


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name="preprocess_import")
def run(t: Task) -> None:
    t.logging_level.set()
    t.pickle_protocol.set()
    tyble = IndividualBasicRow.schema().read_csv(t.root / t.input_table, return_type="Tyble")
    inputs: Sequence[int] = t.inputs

    if not inputs:
        inputs = list(range(len(tyble)))
    rows1: List[IndividualImportedRow] = []
    for i in inputs:
        row = tyble[i]
        input_filename = t.root / t.input_folder / row.raw_filename
        output_filename = t.output_pattern.to_path(t.root, row.name)
        data = fits.getdata(input_filename)
        header = fits.getheader(input_filename)
        output_pickle, output_row = preprocess_import(
            row=row,
            header=header,
            data=data,
            instrument=t.instrument,
            plx_mas=t.plx_mas,
            drs_style=t.drs_style,
        )
        rows1.append(output_row)
        save_pickle(output_filename, output_pickle)

    output_table = t.root / t.output_table
    output_table_lockfile = output_table.with_suffix(output_table.suffix + ".lock")

    logging.debug(f"Appending to output table {output_table}")
    with FileLock(output_table_lockfile):
        df = IndividualImportedRow.schema().from_rows(rows1, return_type="DataFrame")
        df.to_csv(output_table, header=not output_table.exists(), mode="a", index=False)


def cli() -> None:
    run(Task.from_command_line_())



from typing import Literal, NamedTuple, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.stats import norm
from typing_extensions import TypeAlias, assert_never

absurd_minus_99_9: float = -99.9

Float: TypeAlias = Union[float, np.float64]

c_lum = 299.792e6
h_planck = 6.626e-34
k_boltz = 1.3806e-23


def create_grid(wave_min: Float, dwave: Float, nb_bins: int) -> NDArray[np.float64]:
    """
    Create an equidistant wavelength vector

    Args:
        wave_min: Minimum wavelength value
        dwave: Wavelength step
        nb_bins: Length of the wavelength vector

    Returns:
        The vector containing the equidistant values.
    """

    return np.linspace(
        wave_min, wave_min + (nb_bins - 1) * dwave, nb_bins, dtype=np.float64
    )  # the grid of wavelength of your spectrum (assumed equidistant in lambda)


class DopplerResult(NamedTuple):
    """Result of the doppler_r function"""

    #: With positive shift
    plus_v: NDArray[np.float64]

    #: With negative shift
    minus_v: NDArray[np.float64]


def doppler_r(
    lamb: NDArray[np.float64],
    v: Float,
) -> DopplerResult:
    """
    Relativistic doppler shift of a wavelength by a velocity v in kms.

    Args:
        lamb: Wavelength
        v: Velocity in km/s

    Returns:
        A tuple of shifted wavelengths
    """

    factor = np.sqrt((1 + 1000 * v / c_lum) / (1 - 1000 * v / c_lum))
    lambo = lamb * factor
    lambs = lamb * factor ** (-1)
    return DopplerResult(lambo, lambs)


def gaussian(x, cen, amp, offset, wid):
    """Gaussian function"""
    return amp * np.exp(-0.5 * (x - cen) ** 2 / wid**2) + offset


def local_max(
    spectre: NDArray[np.float64], vicinity: int
) -> Tuple[NDArray[np.int64], NDArray[np.float64]]:
    """
    Perform a local maxima algorithm of a vector.

    Args:
        spectre: The vector to investigate.
        vicinity: The half window in which a local maxima is searched.

    Returns:
        A tuple containing the index and vector values of the detected local maxima.
    """

    vec_base = spectre[vicinity:-vicinity]
    maxima = np.ones(len(vec_base))
    for k in range(1, vicinity):
        maxima *= (
            0.5
            * (1 + np.sign(vec_base - spectre[vicinity - k : -vicinity - k]))
            * 0.5
            * (1 + np.sign(vec_base - spectre[vicinity + k : -vicinity + k]))
        )

    index = np.where(maxima == 1)[0] + vicinity
    flux = spectre[index]
    return (index, flux)


def smooth(
    y: NDArray[np.float64],
    box_pts: float,
    shape: Literal["rectangular", "gaussian", "savgol"] = "rectangular",
) -> NDArray[np.float64]:  # rectangular kernel for the smoothing

    """
    Smoothing function.

    Args:
        y: The data to be smoothed.
        box_pts: Half-width of the smoothing window.
        shape: The kernel used to smooth the data

    Returns:
        The smoothed data.

    """

    box2_pts = int(2 * box_pts - 1)
    if shape == "savgol":
        if box2_pts >= 5:
            y_smooth = savgol_filter(y, box2_pts, 3)
        else:
            y_smooth = y
    else:
        if shape == "rectangular":
            box = np.ones(box2_pts) / box2_pts
        elif shape == "gaussian":
            vec = np.arange(-25, 26)
            box = norm.pdf(vec, scale=(box2_pts - 0.99) / 2.35) / np.sum(
                norm.pdf(vec, scale=(box2_pts - 0.99) / 2.35)
            )
        else:
            assert_never(shape)
        y_smooth = np.convolve(y, box, mode="same")
        y_smooth[0 : int((len(box) - 1) / 2)] = y[0 : int((len(box) - 1) / 2)]
        y_smooth[-int((len(box) - 1) / 2) :] = y[-int((len(box) - 1) / 2) :]
    return y_smooth


def truncated(array, spectre, treshold=5):
    maxi = np.percentile(spectre, 99.9)
    mini = np.percentile(spectre, 0.1)
    tresh = (maxi - mini) / treshold
    array[array < mini - tresh] = mini
    array[array > maxi + tresh] = maxi
    return array


def check_none_negative_values(array: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Replace negative numbers an a 1-D vector by the local closest non-zero value.

    Args:
        array: The wavelength vector of the spectrum.

    Returns:
        The vector with null values replaced.

    """
    neg = np.where(array <= 0)[0]
    if len(neg) == 0:
        pass
    elif len(neg) == 1:
        array[neg] = 0.5 * (array[neg + 1] + array[neg - 1])
    else:
        where = np.where(np.diff(neg) != 1)[0]
        if (len(where) == 0) & (np.mean(neg) < len(array) / 2):
            array[neg] = array[neg[-1] + 1]
        elif (len(where) == 0) & (np.mean(neg) >= len(array) / 2):
            array[neg] = array[neg[0] - 1]
        else:
            where = np.hstack([where, np.array([0, len(neg) - 1])])
            where = where[where.argsort()]
            for j in range(len(where) - 1):
                if np.mean(array[neg[where[j]] : neg[where[j + 1]] + 1]) < len(array) / 2:
                    array[neg[where[j]] : neg[where[j + 1]] + 1] = array[neg[where[j + 1]] + 1]
                else:
                    array[neg[where[j]] : neg[where[j + 1]] + 1] = array[neg[where[j]] - 1]
    return array


def make_continuum(
    wave,
    flux,
    flux_denoised,
    grid,
    spectrei,
    continuum_to_produce: Tuple[
        Literal["all", "linear", "cubic"], Literal["all", "denoised", "undenoised"]
    ] = ("all", "all"),
):
    """
    Perform the classical sanity check sequence of continuum.

    Parameters
    ----------
    wave : array_like
        Wavelength vector.
    flux : array_like
        Flux vector.
    flux_denoised : array_like
        Flux vector smoothed.
    grid : array_like
        Length of the window for the savgol filtering.
    spectrei : array_like
        No more used ?
    continuum_to_produce : list
        Specifying on which continuum to perform the sanity check.

    Returns
    -------

    continuum_linear : array_like
        The linearly interpolated continuum
    continuum_cubic : array_like
        The cubicly interpolated continuum
    continuum_linear_denoised : array_like
        The linearly interpolated continuum denoised
    continuum_cubic_denoised : array_like
        The cubicly interpolated continuum denoised

    """

    continuum1_denoised = np.zeros(len(grid))
    continuum3_denoised = np.zeros(len(grid))
    continuum1 = np.zeros(len(grid))
    continuum3 = np.zeros(len(grid))

    if continuum_to_produce[1] != "undenoised":
        if continuum_to_produce[0] != "cubic":
            Interpol1 = interp1d(
                wave, flux_denoised, kind="linear", bounds_error=False, fill_value="extrapolate"
            )
            continuum1_denoised = Interpol1(grid)
            continuum1_denoised = truncated(continuum1_denoised, spectrei)
            continuum1_denoised = check_none_negative_values(continuum1_denoised)
        if continuum_to_produce[0] != "linear":
            Interpol3 = interp1d(
                wave, flux_denoised, kind="cubic", bounds_error=False, fill_value="extrapolate"
            )
            continuum3_denoised = Interpol3(grid)
            continuum3_denoised = truncated(continuum3_denoised, spectrei)
            continuum3_denoised = check_none_negative_values(continuum3_denoised)

    if continuum_to_produce[1] != "denoised":
        if continuum_to_produce[0] != "cubic":
            Interpol1 = interp1d(
                wave, flux, kind="linear", bounds_error=False, fill_value="extrapolate"
            )
            continuum1 = Interpol1(grid)
            continuum1 = truncated(continuum1, spectrei)
            continuum1 = check_none_negative_values(continuum1)
        if continuum_to_produce[0] != "linear":
            Interpol3 = interp1d(
                wave, flux, kind="cubic", bounds_error=False, fill_value="extrapolate"
            )
            continuum3 = Interpol3(grid)
            continuum3 = truncated(continuum3, spectrei)
            continuum3 = check_none_negative_values(continuum3)

    return continuum1, continuum3, continuum1_denoised, continuum3_denoised



from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Sequence, TypedDict

import configpile as cp
import numpy as np
import numpy.typing as npt
import tybles as tb
from typing_extensions import Annotated

from ..lib.analysis import find_nearest1
from ..lib.data import LoggingLevel, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.math import create_grid
from ..lib.util import log_task_name_and_time
from .types import MasterPickle, MasterRow, StackedBasicRow, StackedPickle


@dataclass(frozen=True)
class Task(cp.Config):
    """Creates a master spectrum from stacked spectra"""

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Pickle protocol version to use
    pickle_protocol: Annotated[
        PickleProtocol, cp.Param.store(PickleProtocol.parser(), default_value="3")
    ]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #

    prog_ = Path(__file__).stem

    ini_strict_sections_ = [Path(__file__).stem.split("_")[0]]

    #: Input stacked basic table
    input_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-I")]

    #: Output master spectrum info table (one row)
    output_table: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-O")]

    #: Folder containing the stacked spectra
    input_folder: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-i")]

    #: Path to the master spectrum file to write
    output_file: Annotated[Path, cp.Param.store(cp.parsers.path_parser, short_flag_name="-o")]

    #: Group indices to process
    #:
    #: If not provided, all groups are processed
    groups: Annotated[
        Sequence[int],
        cp.Param.append1(
            cp.parsers.int_parser,
            positional=cp.Positional.ZERO_OR_MORE,
            long_flag_name=None,
            short_flag_name=None,
        ),
    ]


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    t.pickle_protocol.set()
    (t.root / t.output_table).parent.mkdir(parents=True, exist_ok=True)
    (t.root / t.output_file).parent.mkdir(parents=True, exist_ok=True)

    logging.debug(f"Reading {t.root/t.input_table}")
    input_tyble = StackedBasicRow.schema().read_csv(t.root / t.input_table, return_type="Tyble")
    first = input_tyble[0]
    nb_bins = first.nb_bins
    stack = np.zeros(nb_bins, dtype=np.float64)
    berv_mins = [r.berv_min for r in input_tyble]
    berv_maxs = [r.berv_max for r in input_tyble]
    nb_spectra = [r.nb_spectra_stacked for r in input_tyble]
    all_berv = np.array([r.berv for r in input_tyble])
    all_snr = np.array([r.SNR_5500 for r in input_tyble])
    wave_min = input_tyble[0].wave_min
    dwave = input_tyble[0].dwave
    nb_bins = input_tyble[0].nb_bins
    grid = create_grid(wave_min, dwave, int(nb_bins))

    for row in input_tyble:
        data = open_pickle(t.root / t.input_folder / f"{row.name}.p", StackedPickle)
        stack += data["flux"]

    stack[stack <= 0.0] = 0.0
    wave_ref = int(find_nearest1(grid, 5500)[0])
    continuum_5500 = np.nanpercentile(stack[wave_ref - 50 : wave_ref + 50], 95)
    SNR = np.sqrt(continuum_5500)
    BERV = np.sum(all_berv * all_snr**2) / np.sum(all_snr**2)
    BERV_MIN = np.min(berv_mins)
    BERV_MAX = np.max(berv_maxs)
    out: MasterPickle = {
        "flux": stack,
        "master_spectrum": True,
        "RV_sys": first.rv_mean,
        "RV_shift": np.float64(0.0),
        "SNR_5500": SNR,
        "lamp_offset": np.float64(0.0),
        "acc_sec": first.acc_sec,
        "berv": BERV,
        "berv_min": BERV_MIN,
        "berv_max": BERV_MAX,
        "instrument": first.instrument,
        "mjd": np.float64(0.0),
        "jdb": np.float64(0.0),
        "hole_left": first.hole_left,
        "hole_right": first.hole_right,
        "wave_min": first.wave_min,
        "wave_max": first.wave_max,
        "dwave": first.dwave,
        "nb_spectra_stacked": int(np.sum(nb_spectra)),
        "arcfiles": "none",
    }
    save_pickle(t.root / t.output_file, out)
    out_row = MasterRow(
        name=t.output_file.stem,
        SNR_5500=SNR,
        acc_sec=first.acc_sec,
        berv=BERV,
        berv_min=BERV_MIN,
        berv_max=BERV_MAX,
        instrument=first.instrument,
        hole_left=first.hole_left,
        hole_right=first.hole_right,
        wave_min=first.wave_min,
        wave_max=first.wave_max,
        dwave=first.dwave,
        nb_spectra_stacked=int(np.sum(nb_spectra)),
    )
    MasterRow.schema().from_rows([out_row]).to_csv(t.root / t.output_table, index=False)


def cli() -> None:
    run(Task.from_command_line_())



import argparse
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import List, Sequence

import configpile as cp
import pandas as pd
from typing_extensions import Annotated

from ..lib.data import LoggingLevel
from ..lib.util import log_task_name_and_time


@dataclass(frozen=True)
class Task(cp.Config):
    """
    Sorts the rows of a CSV file according to a specified column
    """

    #
    # Common information
    #

    env_prefix_ = "RASSINE"

    #: Use the specified configuration files.
    #:
    #: Files can be separated by commas/the command can be invoked multiple times.
    config: Annotated[Sequence[Path], cp.Param.config(env_var_name="RASSINE_CONFIG")]

    #: Root path of the data, used as a base for other relative paths
    root: Annotated[Path, cp.Param.store(cp.parsers.path_parser, env_var_name="RASSINE_ROOT")]

    #: Logging level to use
    logging_level: Annotated[
        LoggingLevel,
        cp.Param.store(
            LoggingLevel.parser(), default_value="WARNING", env_var_name="RASSINE_LOGGING_LEVEL"
        ),
    ]

    #
    # Task specific information
    #
    prog_ = Path(__file__).stem

    #: Column to sort by
    column: Annotated[str, cp.Param.store(cp.parsers.stripped_str_parser, short_flag_name="-c")]

    #: CSV file to reorder
    file: Annotated[
        Path,
        cp.Param.store(
            cp.parsers.path_parser,
            short_flag_name=None,
            long_flag_name=None,
            positional=cp.Positional.ONCE,
        ),
    ]


def get_parser() -> argparse.ArgumentParser:
    """Returns the argument parser for Sphinx doc purposes"""
    return Task.get_argument_parser_()


@log_task_name_and_time(name=Path(__file__).stem)
def run(t: Task) -> None:
    t.logging_level.set()
    data = pd.read_csv(t.root / t.file)
    assert t.column in data.columns
    data.sort_values(t.column, inplace=True)

    # reread file lines
    with open(t.root / t.file, "r") as f:
        lines = f.readlines()

    header = lines[0]
    output_lines: List[str] = [header, *[lines[i + 1] for i in data.index]]
    with open(t.root / t.file, "w") as f:
        f.writelines(output_lines)


def cli() -> None:
    run(Task.from_command_line_())



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional, Sequence, TypedDict, Union

import numpy as np
from numpy.typing import NDArray

from ..lib.math import Float


class RassineBasicOutput(TypedDict):
    continuum_linear: NDArray[np.float64]
    anchor_wave: NDArray[np.float64]
    anchor_flux: NDArray[np.float64]
    anchor_index: NDArray[np.int64]


class RassineParameters(TypedDict):
    filename: str
    number_iteration: int
    K_factors: Sequence[float]
    axes_stretching: Float
    vicinity_local_max: int
    smoothing_box: Union[Literal["auto"], int]
    smoothing_kernel: Literal["rectangular", "gaussian", "savgol", "erf", "hat_exp"]
    fwhm_ccf: Float
    CCF_mask: str
    RV_sys: Float
    min_radius: Float
    max_radius: Float
    model_penality_radius: str
    denoising_dist: int
    number_of_cut: int
    windows_penality: Float
    large_window_penality: Float
    number_points: int
    number_anchors: int
    SNR_5500: int
    mjd: Float
    jdb: Float
    wave_min: Float
    wave_max: Float
    dwave: Float
    hole_left: Float
    hole_right: Float
    RV_shift: Float
    berv: Float
    lamp_offset: Float
    acc_sec: Float
    light_file: Literal[True]
    speedup: Literal[1]
    continuum_interpolated_saved: Literal["linear"]
    continuum_denoised_saved: Literal["undenoised"]
    nb_spectra_stacked: int
    arcfiles: Sequence[str]

    #: Time weighted by the radial velocity uncertainty (info for YARARA)
    #:
    #: Not present in the master spectrum
    rv_mean_jdb: Optional[np.float64]

    #: Std deviation weighted radial velocity average (info for YARARA)
    #:
    #: Comes from the stacking_basic table mean_vrad column (info for YARARA)
    #:
    #: Not present in the master spectrum
    rv_dace: Optional[np.float64]

    #: Radial velocity standard deviation
    #:
    #: Comes from the stacking_basic table mean_svrad column
    #:
    #: Not present in the master spectrum
    rv_dace_std: Optional[np.float64]


class RassinePickle(TypedDict):
    #: Wavelength
    wave: NDArray[np.float64]
    #: Initial flux
    flux: NDArray[np.float64]
    #: Initial flux error, passed through by RASSINE
    flux_err: Optional[NDArray[np.float64]]
    #: Smoothed flux
    flux_used: NDArray[np.float64]
    #: Rassine output continuum
    output: RassineBasicOutput
    #: Rassine derived parameters
    parameters: RassineParameters


@dataclass(frozen=True)
class ExtraPlotData:
    normalisation: float
    spectre: NDArray[np.float64]
    conti: NDArray[np.float64]
